import threading
import requests, os, subprocess, time, logging, struct, signal, re
import socket, ipaddress, urllib3, binascii, json, atexit, sys, traceback
from datetime import datetime
from functools import partial
from urllib3.exceptions import InsecureRequestWarning
from os import path
from requests.auth import HTTPBasicAuth
from pathlib import Path
import webbrowser
import logging.handlers

import cnvpnclient.files as file_utils
from cnvpnclient import os_utils, wireguard
from cnvpnclient.exception import ClientException
from cnvpnclient.logutil import json_stringify, configure_client_logger
from cnvpnclient.config import LibConfig, ConnectionTypes

lib_logger = logging.getLogger("cnvpnclient")

requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)

"""
Documentation on props structure
props {
  # These props are parsed or calculated from the clientpack directly. cn=cohesive specific
  Endpoint: str, endpoint for vns3 controller to connect to 
  TunnelAllTraffic: bool, add route for 0.0.0.0/0 to send all internet traffic through vpn [cn]
  SecondaryEndpoints: str, csv of secondary endpoints for failover [cn]
  RoutePolling: bool, if true, will poll VNS3 for routes [cn]
  PresharedKey: str, wg property
  ActiveInterface: str, actual interface up on OS. different based on OS.
  PublicKey: str, wg property
  PrivateKey: str, wg property
  Secret: str, API token to use for sending requests to VNS3 [cn]
  ClientpackId: str, VNS3 clientpack ID [cn]
  Address: str, wg property
  AllowedIPs: str, csv, wg property
  LocalRoutes: str, csv - local routes to add for VPN overlay network
  RoutePollingInterval: int, time (seconds) to wait between polling routes
  HandshakeTimeout: int, time (seconds) that defines a timeout for the connection
  ConnectionStatus: str, store state of connection. empty string or 'up"
  
  overlay_ip: str, this is the Address of the conf and is this clients IP
  interface: str, interface requested (not necessarily the ActiveInterface)
  wg_conf_file: str, path to wireguard conf file
  wg_auth: str, type of auth, not sure we use this anymore
  controller_vip: str, fetched from VNS3 when connecting
  openid: str, state string indicating if openid auth configured on server
  token: str, retrieved after openid login
  refresh_token: str, retrieved after openid login
  routes_file: str, path to file to store routes in
  type: str, ConnectionTypes. wg-native wg-cohesive
  name: str, name of connection
  state_file: str, path to state file
  _auth_attempts: int, state tracker for openid auth attempts
}
"""


def get_current_endpoint(props):
    _endpoint = props.get("Endpoint")
    if not _endpoint:
        return None
    return _endpoint.split(":")[0]


def update_props_endpoint(props, new_endpoint):
    endpoint_parts = props["Endpoint"].split(":")
    port = endpoint_parts[1] if len(endpoint_parts) > 1 else ""
    update = "%s:%s" % (new_endpoint, port) if port else new_endpoint
    props.update(Endpoint=update)
    # mutates and returns
    return props


def get_secondary_endpoints_props(props):
    """get_secondary_endpoints_props

    Parse and clean SecondaryEndpoints from props dict

    Args:
        props (dict)

    Returns: list[str]
    """
    secondary_endpoints = props.get("SecondaryEndpoints", "")
    endpoints_cln = [x.strip() for x in secondary_endpoints.split(",")]
    # filter empty strings
    return [endpoint for endpoint in endpoints_cln if endpoint]


# fetch the connection info for the clientpack
# thats needed to make the wireguard connection
def wg_connect(props, clientpack_id, api_token, logger=None, timeout=None):
    """wg_connect

        fetch the connection info for the clientpack
        thats needed to make the wireguard connection
    Args:
        props (dict): client properties
        clientpack_id (str): id/name of clientpack
        api_token (str): API token
        logger (optional): Defaults to library logger.

    Returns: 200 indicating success or None
    """
    logger = logger or lib_logger
    controller = get_current_endpoint(props)
    logger.debug(
        f"Validating connection on {controller} for clientpack {clientpack_id}"
    )
    my_headers = {
        "Content-Type": "application/json",
        "api-token": api_token,
        "state": props["openid"],
        "token": props["token"],
        "refresh-token": props["refresh_token"],
    }
    try:
        response = requests.get(
            f"https://{controller}:8000/api/wg-connection/{clientpack_id}",
            headers=my_headers,
            verify=False,
            timeout=(timeout or LibConfig.DEFAULT_CONNECT_TIMEOUT),
        )
    except Exception as e:
        logger.info("Failed to call vns3 connect: %s" % str(e))
        return None

    if "DOCTYPE".lower() in response.text.lower():
        props["_auth_attempts"] = props.get("_auth_attempts", 0) + 1
        logger.debug("Waiting on authentication. Attempt %s" % props["_auth_attempts"])
        if props["_auth_attempts"] > 20:  # should we have this?
            raise ClientException(
                "Failed waiting for login via browser. Please retry and login via the browser pop-up window"
            )
        return None

    if response.status_code == 401:
        try:
            err_resp = response.json()
        except json.JSONDecodeError:
            err_resp = None

        unauth_message = err_resp.get("error", {}).get("message") if err_resp else ""
        # if not unauth_message or unauth_message == "VPNCubed::API::UnauthorizedError":
        if unauth_message.lower() != "user must login":
            raise ClientException(
                "Failed to connect. Clientpack is unauthorized. Please review the clientpack Secret"
            )

        props["openid"] = binascii.b2a_hex(os.urandom(16)).decode("UTF-8")
        logger.debug(f"Received unauthorized, user needs to login")
        props["token"] = ""
        props["refresh_token"] = ""
        webbrowser.open(
            f"https://{controller}:8000/api/wg-connection/{clientpack_id}?state={props['openid']}",
            new=2,
        )
        return None

    elif response.status_code == 200:
        if response.text == "null":
            logger.debug(
                "received empty response validating wireguard connection with controller, ignore"
            )
            return None
        logger.debug(
            f"Validated wireguard connection with controller, now lets connect"
        )
        resp = json.loads(response.text)
        psk = resp["response"]["psk"]
        controller_vip = resp["response"]["controller_vip"]
        refresh_token = resp["response"]["refresh_token"]
        if controller_vip != "":
            props["controller_vip"] = controller_vip
        if refresh_token != "":
            props["refresh_token"] = refresh_token
        if psk != "":
            props["PresharedKey"] = psk
        else:
            props.pop("PresharedKey", None)

        props["wg_auth"] = json.loads(response.text)["response"]["auth"]
        return 200

    else:
        error_msg = "connect failure: %s" % response.text
        logger.info(error_msg)
        raise ClientException(error_msg)


def wg_disconnect(endpoint, clientpack_id, api_token, logger=None, timeout=None):
    logger = logger or lib_logger
    logger.debug(f"Disconnecting from {endpoint} for clientpack {clientpack_id}")
    my_headers = {"api-token": api_token}
    try:
        response = requests.put(
            f"https://{endpoint}:8000/api/wg-connection/{clientpack_id}",
            headers=my_headers,
            verify=False,
            timeout=(timeout or LibConfig.DEFAULT_CONNECT_TIMEOUT),
        )
    except Exception as e:
        logger.error(str(e))
        return

    if response.status_code == 200:
        logger.debug("Disconnecting acknowledged %s" % clientpack_id)
    else:
        # Question better error message
        logger.info(
            f"Unexpected status code disconnecting [{response.status_code}]: {response.text}"
        )


def run_command_windows(command, logger=None, should_raise=False):
    process_kwargs = {
        "stdout": subprocess.PIPE,
        "stderr": subprocess.PIPE,
        "shell": False,
        "creationflags": subprocess.CREATE_NO_WINDOW
    }

    with subprocess.Popen(command, **process_kwargs) as proc:
        output, error = proc.communicate()
        if output is not None:
            output = output.decode("utf-8").strip()
        if error is not None:
            error = error.decode("utf-8").strip()

        if error:
            error_str = "Command stderr. command=%s: %s" % (command, error)
            if logger:
                logger.debug(error_str)
            proc.terminate()
            if should_raise:
                raise ClientException(error_str)
            return (output, error)
        return (output, error)


# utility method to run a system command
def run_command(command, logger=None, should_raise=False):
    process_kwargs = {
        "stdout": subprocess.PIPE,
        "stderr": subprocess.PIPE,
        "shell": True,
    }
    with subprocess.Popen(command, **process_kwargs) as proc:
        output, error = proc.communicate()
        if output is not None:
            output = output.decode("utf-8").strip()
        if error is not None:
            error = error.decode("utf-8").strip()

        if error:
            error_str = "Command stderr. command=%s: %s" % (command, error)
            if logger:
                logger.debug(error_str)
            proc.terminate()
            if should_raise:
                raise ClientException(error_str)
            return (output, error)
        return (output, error)


# return the last handshake time using wireguard system call
# def get_last_handshake(interface):
#     return run_command(
#         "wg show " + interface + " latest-handshakes | head -n1 | awk '{print $2;}'"
#     )


def get_last_handshake(env, interface):
    show_latest_handshakes = "%s show %s latest-handshakes" % (
        env.get("wg_path"),
        interface,
    )
    if env.is_windows:
        cmd1 = ["powershell", show_latest_handshakes]
        latest_handshakes, _ = run_command(cmd1)
        parse_cmd = "echo '%s'| %%{ $_.Split('=')[1]; }" % latest_handshakes
        command = ["powershell", parse_cmd]
        out, err = run_command_windows(command)    
    else:
        command = "%s | head -n1 | awk '{print $2;}'" % show_latest_handshakes
        out, err = run_command(command)

    if err and "permission" in err.lower():
        raise ClientException(
            "Invalid permissions for client. Cannot query OS interfaces. Requires root/admin permissions."
        )
    return out


# return the date from a specified number of seconds ago
def get_check_date(env, timeout):
    if env.is_linux:
        out, err = run_command(f"date -d '-{timeout} seconds' '+%s'")
    elif env.is_macos:
        out, err = run_command(f"date -v -{timeout}S +%s")
    elif env.is_windows:
        out, err = run_command_windows(
            ["powershell", f"(Get-Date).AddSeconds(-{timeout}) | Get-Date -UFormat %s"]
        )
    else:
        raise ClientException("Unknown platform %s" % env.platform)
    return out


# if the last handshake is over the check date, then we need to reset interace
def needs_interface_reset(env, props, handshake_timeout_interval):
    if not props["ActiveInterface"]:
        if env.state_file:
            file_utils.put_key_value(env.state_file, "status", "connecting")
        return True

    last_handshake = get_last_handshake(env, props["ActiveInterface"])
    if not last_handshake:
        env.logger.info("Client is disconnected. No handshake detected.")
        props["ConnectionStatus"] = ""
        if env.state_file:
            file_utils.put_key_value(env.state_file, "status", "connecting")
        return True

    if get_last_handshake(env, props["ActiveInterface"]) > get_check_date(
        env, handshake_timeout_interval
    ):
        current_status = props["ConnectionStatus"]
        props["ConnectionStatus"] = "up"
        if current_status != "up" and env.state_file:
            file_utils.update_data_file(
                env.state_file,
                {
                    "status": "connected",
                    "pid": os.getpid(),
                    "last_connect": datetime.utcnow().strftime(
                        LibConfig.DATETIME_FORMAT
                    ),
                },
            )
        return False

    env.logger.info(
        f"No connection handshake in last {handshake_timeout_interval} seconds for {props['ActiveInterface']}"
    )
    props["ConnectionStatus"] = ""
    if env.state_file:
        file_utils.put_key_value(env.state_file, "status", "connecting")
    return True


# update the preshared key in the wg config file
def update_preshared_key_config(wg_conf_file, psk):
    wg_fin = open(wg_conf_file, "r")
    prev_psk = ""
    config = ""
    for line in wg_fin:
        if line.startswith("PresharedKey"):
            if psk == "":
                line = ""
            else:
                prev_psk = line.partition(" =")[2].strip()
                line = line.replace(prev_psk, psk)
        config = config + line
    if psk != "" and config.find("PresharedKey") == -1:
        config = config.replace(f"[Peer]", f"[Peer]\nPresharedKey = {psk}")
    wg_fout = open(wg_conf_file, "w")
    with open(wg_conf_file, "w") as wg_fout:
        wg_fout.write(config)


# wg-quick with run_command returns error when it may not be
# for now, if there is no "error" in the message, assume no error
# would be good to think of a better way to handle this
def handle_wg_quick_output(stdout, err, logger=None):
    if err and "error" in err.lower():
        logger.info(f"wg-quick error: {err}")
        raise ClientException(err)
    return (stdout, None)


def display_interface_status(env, interface, timeout):
    logger = env.logger
    if not interface:
        return

    if wireguard.is_interface_down(env, interface):
        logger.debug(f"{interface} interface is down")
        return

    lhs = get_last_handshake(env, interface)
    if lhs == "0" or lhs == "":
        logger.info(f"{interface} must be down, no handshake {lhs}")
    else:
        if env.is_windows:
            check_hs, err = run_command_windows(
                [
                    "powershell",
                    f"{env.get('wg_path')} show {interface} | Select-String -Pattern 'latest handshake'"
                ]
            )
        else:
            check_hs, err = run_command(
                f"{env.get('wg_path')} show {interface} | grep 'latest handshake'"
            )
        logger.debug(
            f"Wireguard {interface} is up, {check_hs} (reset if no handshake within: {timeout} seconds)"
        )


def parse_clientpack(clientpack_path_or_lines):
    """parse_clientpack

    Parse file for clientpack configuration details. Currently only wiregaurd
    is supported

    Args:
        clientpack_path_or_lines (str|List[str]): path to clientpack file or list of clientpack lines

    Returns:
        dict: data
    """
    data = {}
    if type(clientpack_path_or_lines) is str:
        with open(clientpack_path_or_lines, "r") as cp_file:
            clientpack_lines = cp_file.readlines()
        data["wg_cp_file"] = clientpack_path_or_lines
        data["interface"] = os.path.basename(clientpack_path_or_lines).split(".")[0]
    else:
        clientpack_lines = clientpack_path_or_lines

    for line in clientpack_lines:
        if "=" in line:
            toks = line.partition("=")
            token = toks[0].strip()
            if token == "#Endpoint" or token == "#MTU" or token == "#DNS":
                continue  # ignore these tokens
            if token.startswith("#"):
                token = token[1:]
            data[token] = toks[2].strip()

    data["openid"] = ""
    data["token"] = ""
    data["refresh_token"] = ""
    data["ConnectionStatus"] = ""

    if "Address" in data:
        data["ClientpackId"] = data["Address"].rpartition("/")[0].replace(".", "_")

    secondary_endpoints = data.get("SecondaryEndpoints")
    endpoint = data.get("Endpoint")
    if secondary_endpoints and endpoint:
        data["SecondaryEndpoints"] = endpoint.split(":")[0] + ", " + secondary_endpoints
    elif endpoint:
        data["SecondaryEndpoints"] = endpoint.split(":")[0]

    return data


def wg_allowed_ips(env, routes, props):
    if routes is None:
        routes = []
    append_routes = ""
    for r in set(routes):
        append_routes = append_routes + ", " + r[0] + "/" + str(netmask_to_cidr(r[1]))

    set_cmd = f"{env.get('wg_path')} set {props['ActiveInterface']} peer {props['PublicKey']} allowed-ips \"{props['AllowedIPs']}{append_routes}\""
    if env.is_windows:
        command = ["powershell", set_cmd]
        out, err = run_command_windows(command)
    else:
        command = set_cmd
        out, err = run_command(command)
    return out


def ping_controller_via_overlay(env, controller, clientpack, timeout=None):
    try:
        response = requests.get(
            f"https://{controller}:8000/api/status/connected_subnets?cp={clientpack}",
            headers={},
            auth=HTTPBasicAuth("overlayapi", "x"),
            verify=False,
            timeout=(timeout or LibConfig.DEFAULT_CONNECT_TIMEOUT),
        )
        return response.status_code == 200
    except Exception as e:
        env.logger.info(
            "Failed to ping controller %s via overlay. Error=%s" % (controller, str(e))
        )
        return False


def update_routes(env, props, routes_file, server, interface, cp, src_cidr):
    src_address = src_cidr.split("/")[0]
    new_routes = vpn3api_desc_connected_subnets(env, server, cp)
    routes = []
    for route in open(routes_file).readlines():
        a, b = route.replace("(", "").replace(")", "").split(",")
        routes.append((a.strip().replace("'", ""), b.strip().replace("'", "")))

    if new_routes is None:
        new_routes = []
    if new_routes != routes:
        routes_to_add = set(new_routes).difference(routes)
        routes_to_del = set(routes).difference(new_routes)
        env.logger.info(
            "Route change detected. Add=%s Del=%s" % (routes_to_add, routes_to_del)
        )
        wg_allowed_ips(env, new_routes, props)
        for r in routes_to_add:
            add_route(env, r, interface, src_address)
        for r in routes_to_del:
            del_route(env, r, interface)
        routes = new_routes

    with open(routes_file, "w") as outfile:
        outfile.write("\n".join(str(item) for item in new_routes))


def delete_dynamic_routes(env, props):
    routes_file = props.get("routes_file")
    if routes_file:
        for route in open(routes_file).readlines():
            a, b = route.replace("(", "").replace(")", "").split(",")
            route = (
                a.strip().replace("'", ""),
                b.strip().replace("'", ""),
            )
            # Question can be interface from props?
            del_route(env, route, props["ActiveInterface"])


def delete_all_routes(env, props):
    delete_dynamic_routes(env, props)
    delete_local_routes(env, props)
    if tunnel_all_traffic(props):
        delete_endpoint_routes(env, props)


def vpn3api_desc_connected_subnets(env, controller, cp, timeout=None):
    """
    Returns None when an error was encountered; returns
    a sorted list of route tuples otherwise
    """
    try:
        env.logger.debug("fetching routes from %s" % controller)
        response = requests.get(
            f"https://{controller}:8000/api/status/connected_subnets?cp={cp}",
            headers={},
            auth=HTTPBasicAuth("overlayapi", "x"),
            verify=False,
            timeout=(timeout or LibConfig.DEFAULT_CONNECT_TIMEOUT),
        )
        if response.status_code != 200:
            env.logger.error(
                "Fetching routes failed (%s): %s"
                % (response.status_code, response.text)
            )
            return None

        routes = response.json().get("response")
        env.logger.debug("Fetched routes: %s" % routes)
        return sorted([tuple(r) for r in routes])
    except Exception as e:
        env.logger.error(
            "Fetching routes failed. Controller=%s Exception=%s" % (controller, str(e))
        )
        return None


def netmask_to_cidr(m_netmask):
    return sum([bin(int(bits)).count("1") for bits in m_netmask.split(".")])


def cidr_to_netmask(cidr):
    network, net_bits = cidr.split("/")
    host_bits = 32 - int(net_bits)
    netmask = socket.inet_ntoa(struct.pack("!I", (1 << 32) - (1 << host_bits)))
    return netmask


def add_local_routes(env, props, logger=None):
    if "LocalRoutes" in props and props["LocalRoutes"] != "":
        routes = props["LocalRoutes"].split(",")
        for route in routes:
            (logger or lib_logger).info(f"adding local route: {route}")
            if env.is_linux:
                gateway, err = run_command(
                    f"ip route get {route} | cut -d' ' -f 3 | xargs"
                )
                run_command(f"ip route add {route} via {gateway}")

            elif env.is_macos:
                address = route.split("/")[0]
                cidr = route.split("/")[1]
                netmask = cidr_to_netmask(route)
                gateway, err = run_command(
                    "route -n get default | grep gateway | cut -f2 -d':' | xargs"
                )
                run_command(
                    f"route add -net {route} -gateway {gateway} -netmask {netmask}"
                )

            else:
                address = route.split("/")[0]
                cidr = route.split("/")[1]
                gateway, err = get_default_gateway(env, props)
                run_command(
                    f"route ADD {address} MASK 255.255.255.255 {gateway} metric 1"
                )


def delete_local_routes(env, props):
    if "LocalRoutes" in props and props["LocalRoutes"] != "":
        routes = props["LocalRoutes"].split(",")
        for route in routes:
            env.logger.info(f"removing local route: {route}")
            if env.is_posix:
                run_command(f"route delete {route}")

    secondary_endpoints = get_secondary_endpoints_props(props)
    for route in secondary_endpoints:
        route = os_utils.resolve_ip(route)
        env.logger.info(f"removing local controller route: {route}")
        if env.is_posix:
            run_command(f"route delete {route}")


def get_wg_adapter_index(interface):
    out, err = run_command(
        [
            "powershell",
            "Get-NetAdapter -Name {interface} | select -ExpandProperty ifIndex -first 1".format(interface=interface),
        ]
    )
    return out


def add_controller_routes(env, props):
    def_log_message = (
        f"add local routes for current controller and secondary controllers"
    )
    env.logger.debug(f"{def_log_message}")

    controller = get_current_endpoint(props)
    if controller is None:
        env.logger.info(
            f"{def_log_message} failed: could not determine controller endpoint"
        )
        return

    # find the local gateway address
    gateway, err = get_default_gateway(env, props)
    if err:
        env.logger.info(
            f"{def_log_message} failed: cound not get default gateway: {err}"
        )
        return

    if gateway is None:
        env.logger.info(
            f"{def_log_message} failed: could not determine gateway address"
        )
        return

    for endpoint in get_secondary_endpoints_props(props):
        endpoint = os_utils.resolve_ip(endpoint)
        env.logger.debug(f"adding local route for controller: {endpoint}")
        # find platform specific gateway command
        if env.is_linux:
            command = f"ip route add {endpoint}/32 via {gateway}"

        elif env.is_macos:
            command = f"route add -net {endpoint}/32 {gateway}"

        elif env.is_windows:
            command = f"route ADD {endpoint} MASK 255.255.255.255 {gateway} metric 1"

        else:
            env.logger.info(f"{def_log_message} failed: no platform support")
            return

        # add the controller route
        response, err = run_command(command)
        if err:
            env.logger.info(f"{def_log_message} failed: {err}")
            return

    env.logger.debug(f"{def_log_message}: completed")


def get_default_gateway(env, props):
    def_log_message = "get default gateway"

    controller = get_current_endpoint(props)
    if controller is None:
        return (
            None,
            "{msg} failed: Controller endpoint could not be determined".format(msg=def_log_message),
        )

    if env.is_linux:
        gateway, err = run_command(
            f"ip route get {controller} | cut -d' ' -f 3 | xargs"
        )

    elif env.is_macos:
        gateway, err = run_command(
            "route -n get default | grep gateway | cut -f2 -d':' | xargs"
        )

    elif env.is_windows:
        gateway, err = run_command_windows(
            [
                "powershell",
                '(Get-NetRoute "0.0.0.0/0").NextHop',  # need to use controller ip here ?
            ]
        )

    else:
        return (None, "{msg}: no platform support found".format(msg=def_log_message))

    return (gateway, err)


def delete_endpoint_routes(env, props):
    env.logger.info("removing controller routes")
    for route in get_secondary_endpoints_props(props):
        route = os_utils.resolve_ip(route)
        env.logger.info(f"removing local controller route: {route}")
        if env.is_posix:
            run_command(f"route delete {route}")


def add_route(env, route, interface, src_address):
    env.logger.info(f"Adding route: {route} {time.asctime()}")
    if env.is_posix:
        mask = netmask_to_cidr(route[1])
        if env.is_linux:
            cmd = f"ip route add {route[0]}/{mask} dev {interface} src {src_address}"
        else:
            cmd = f"route add -net {route[0]}/{mask} -interface {interface}"
        os.system(cmd)
    else:
        # gw = get_overlay_gateway_on_windows(interface)
        # if gw is None:
        #     env.logger.info("bad gateway %s", time.asctime())
        #     return
        # cmd = "route add %s mask %s " % route
        # cmd += "%s >NUL 2>&1" % gw
        # os.system(cmd)

        # TODO:changed: src_address should be this parsed from Address
        # gw = props['Address'].rpartition("/")[0]
        gw = src_address
        index = get_wg_adapter_index(interface)
        mask = route[1]
        cmd = f"route ADD {route[0]} MASK {mask} {gw} METRIC 1 IF {index}"
        run_command_windows(cmd)


def del_route(env, route, interface):
    env.logger.info(f"Removing route: {route} {time.asctime()}")
    if env.is_posix:
        mask = netmask_to_cidr(route[1])
        if env.is_linux:
            cmd = f"ip route del {route[0]}/{mask} dev {interface}"
        else:
            cmd = f"route delete -net {route[0]}/{mask} -interface {interface}"
        os.system(cmd)
    else:
        cmd = f"route delete {route[0]}"
        run_command_windows(cmd)


def get_overlay_gateway_on_linux(interface):
    f = os.popen(f"ifconfig {interface}")
    gw = None
    for line in f:
        if "P-t-P" in line:
            parts = line.split()
            for p in parts:
                if "P-t-P" in p:
                    gw = p[6:]
    f.close()
    return gw


def initialize(env, props):
    wg_cp_file = props["wg_cp_file"]
    if path.exists(wg_cp_file) == False:
        error = f"Could not find clientpack file: {wg_cp_file}"
        env.logger.error(error)
        raise ClientException(error)

    # Path(env.WG_CP_DIR).mkdir(exist_ok=True)
    # if wg_cp_file != env.SYSTEM_WG_CP_FILE:
    #     lib_logger.info(
    #         f"Found specfied clientpack, copying to {wg_cp_file} to: {env.SYSTEM_WG_CP_FILE}"
    #     )
    #     from_file = pathlib.Path(wg_cp_file)
    #     to_file = pathlib.Path(env.SYSTEM_WG_CP_FILE)
    #     shutil.copy(from_file, to_file)
    #     props['wg_cp_file'] = env.SYSTEM_WG_CP_FILE

    # if path.exists(env.SYSTEM_WG_CP_FILE) == False:
    #     error = "Clientpack copy failed"
    #     lib_logger.error(error)
    #     raise ClientException(error)

    props["wg_auth"] = env.get("wg_auth")

    routes_file = props.get("routes_file")
    if not routes_file:
        return

    props["routes_file"] = routes_file
    # Keep previous routes fetched to compare if new routes occurred
    env.logger.debug(f"Creating/clearing routes file: {routes_file}")
    if env.is_windows:
        if not os.path.exists(routes_file):
            Path(routes_file).write_text("")
    else:
        Path(routes_file).touch()
    Path(routes_file).write_text("")


def update_wg_config(env, props, logger=None):
    logger = logger or lib_logger
    wg_conf_file = props["wg_conf_file"]

    logger.debug(
        f"Updating wireguard config file: {wg_conf_file} with cohesive functionality if exists"
    )

    # custom cohesive stuff
    tunnel_is_gateway = props.get(
        "TunnelAllTraffic", LibConfig.Defaults.TUNNEL_ALL_TRAFFIC
    )
    has_secondary_endpoints = len(get_secondary_endpoints_props(props)) > 0
    check_tunnel_all_traffic = (
        tunnel_is_gateway
        or "128.0.0.0/1" in props["AllowedIPs"]
        or "0.0.0.0/1" in props["AllowedIPs"]
    )

    has_cohesive_edits = check_tunnel_all_traffic or has_secondary_endpoints
    if has_cohesive_edits:
        wg_fin = open(wg_conf_file, "r")
        final_conf_lines = []

        # if line needs to be overwritten, append and continue.
        for line in wg_fin:
            if line.startswith("AllowedIPs"):
                if tunnel_is_gateway:
                    if "128.0.0.0/1" not in props["AllowedIPs"]:
                        props["AllowedIPs"] = f"128.0.0.0/1, {props['AllowedIPs']}"
                    if "0.0.0.0/1" not in props["AllowedIPs"]:
                        props["AllowedIPs"] = f"0.0.0.0/1, {props['AllowedIPs']}"
                    final_conf_lines.append(f"AllowedIPs = {props['AllowedIPs']}")
                    continue

                else:
                    props["AllowedIPs"] = (
                        props["AllowedIPs"]
                        .replace(",", " ")
                        .replace("0.0.0.0/1", " ")
                        .replace("128.0.0.0/1", " ")
                    )

                    props["AllowedIPs"] = ", ".join(props["AllowedIPs"].split())
                    final_conf_lines.append(f"AllowedIPs = {props['AllowedIPs']}")
                    continue

            elif line.startswith("Endpoint"):
                final_conf_lines.append(f"Endpoint = {props['Endpoint']}")
                continue

            elif line.startswith("SaveConfig") and env.is_windows:
                # SaveConfig not supported on windows
                continue

            elif line.startswith("#SecondaryEndpoints"):
                endpoint = get_current_endpoint(props)
                endpoints = [
                    x.strip()
                    for x in props["SecondaryEndpoints"].split(",")
                    if x.strip() != endpoint
                ]
                secondaryEndpoints = ", ".join(endpoints)
                final_conf_lines.append(f"#SecondaryEndpoints = {secondaryEndpoints}")
                continue

            final_conf_lines.append(line.strip())
        # overwrite with final conf
        with open(wg_conf_file, "w") as wg_fout:
            wg_fout.write("\n".join(final_conf_lines))
            os.chmod(
                wg_conf_file, wireguard.CONF_FILE_CHMOD
            )  # user read/write, wg spits out warning if public access


def reset_interface(env, interface, props, logger=None):
    logger = logger or lib_logger
    logger.info(
        f"Reset wireguard connection on interface {interface or props['ClientpackId']}"
    )
    if props["type"] == ConnectionTypes.WIREGUARD_COHESIVE:
        valid_endpoint, _ = get_first_valid_endpoint(
            get_secondary_endpoints_props(props), logger=logger
        )
        if valid_endpoint and valid_endpoint != get_current_endpoint(props):
            props = update_props_endpoint(props, valid_endpoint)
            update_wg_config(env, props, logger)

        vns3_connect_success = wg_connect(
            props, props["ClientpackId"], props["Secret"], logger=logger
        )
        if not vns3_connect_success:
            return False

        if props["ConnectionStatus"] == "":
            if "PresharedKey" in props:
                update_preshared_key_config(
                    props["wg_conf_file"], props["PresharedKey"]
                )
            else:
                update_preshared_key_config(props["wg_conf_file"], "")

    logger.debug(
        f"Bouncing wg interface {props['ActiveInterface'] or 'N/A'} conf={props['wg_conf_file']}"
    )
    wireguard.wg_quick_down(env, props["ActiveInterface"], props["wg_conf_file"])
    _, up_err = wireguard.wg_quick_up(
        env, props["ActiveInterface"], props["wg_conf_file"]
    )

    if up_err:
        # This could be a fatal error. Should we throw here?
        logger.error("ERROR wg-quick up: %s" % up_err)
        return

    # macos
    if interface is None:
        logger.debug("Searching for active interface for IP %s" % props["overlay_ip"])
        actual_interface = os_utils.find_ip_interface(
            env, props["overlay_ip"], retries=5, sleep=0.5
        )
        if actual_interface:
            logger.debug(
                f"Changing interface {interface or 'N/A'} to {actual_interface}"
            )
            props["ActiveInterface"] = actual_interface
            if env.state_file:
                file_utils.put_key_value(env.state_file, "interface", actual_interface)

    interface_details, err = wireguard.get_interface_details(
        env, env.get("wg_path"), props["ActiveInterface"],
        retry_attempts=(10 if env.is_windows else 1), retry_sleep=1.0
    )
    # allow does not exist due to possible race condition.
    # but fail on path path and permissions errors
    if err and "does not exist" not in err:
        raise ClientException("Failed to get interface details after up: %s" % str(err))
    logger.debug("Interface details: %s" % interface_details)

    props["ConnectionStatus"] = "up"
    logger.info(f"Wireguard connection up on {props['ActiveInterface']}")
    if env.state_file:
        file_utils.update_data_file(
            env.state_file,
            {
                "status": "connected",
                "pid": os.getpid(),
                "last_connect": datetime.utcnow().strftime(LibConfig.DATETIME_FORMAT),
            },
        )

    return True


def validate_clientpack(props, logger=None, validate_endpoints=True):
    """validate_clientpack - this DOES mutate props. Output is a valid clientpack
       dictionary using standard python data types

    Args:
        props (dict): dict parsed from clientpack file
        logger (optional): Defaults to library logger.
        validate_endpoints (boolean): Default: True. If True, it will query endpoints for first available

    Returns:
        dict: Always returns dict. empty if none. otherwise errors keyed by type:
        {
            'wg-native': List[str] # list of errors for wg native type,
            'wg-cohesive': List[str] # list of errors for wg cohesive type,
        }
    """
    logger = logger or lib_logger
    connection_type = props.get("type")

    wg_native_errors = []
    wg_cn_errors = []

    if "Address" not in props:
        wg_native_errors.append(
            f"Address must be specified and be a network cidr in conf file"
        )
    else:
        try:
            ipaddress.ip_network(props["Address"], False)
            props["overlay_ip"] = props["Address"].split("/")[0]
        except ValueError:
            wg_native_errors.append(
                f"Address in conf file is not recognized: {props['Address']}"
            )

    if "PrivateKey" not in props:
        wg_native_errors.append(f"PrivateKey must be specified in conf file")

    if "PublicKey" not in props:
        wg_native_errors.append(f"PublicKey must be specified in conf file")

    if "AllowedIPs" not in props:
        wg_native_errors.append(
            f"AllowedIPs must be specified and be a network cidr in conf file"
        )
    else:
        for route in props["AllowedIPs"].split(","):
            if not re.match(
                wireguard.CIDR_CHECK_REGEX,
                route.strip(),
            ):
                wg_native_errors.append(
                    f"AllowedIPs must be specified in a network cidr format: {route}"
                )

    if "SaveConfig" in props:
        props["SaveConfig"] = string_to_boolean(props["SaveConfig"])
        if (
            props["SaveConfig"] is True
            and connection_type == ConnectionTypes.WIREGUARD_COHESIVE
        ):
            wg_cn_errors.append(
                "SaveConfig=true is not supported for cohesive wireguard connection type. Please use native connection type to set SaveConfig=true"
            )

    if "Secret" not in props:
        wg_cn_errors.append(f"Secret must be specified in conf file")

    if "TunnelAllTraffic" in props:
        tunnel_all_traffic_val = string_to_boolean(props["TunnelAllTraffic"])
        if tunnel_all_traffic_val is None:
            wg_cn_errors.append(
                f"TunnelAllTraffic misconfigured, must be True or False"
            )
        else:
            props["TunnelAllTraffic"] = tunnel_all_traffic_val
            logger.debug(
                f"Tunneling all traffic via wireguard interface is {tunnel_all_traffic_val}"
            )
    else:
        props["TunnelAllTraffic"] = False
        logger.debug(
            f"TunnelAllTraffic not configured, can be True or False, setting to '{props['TunnelAllTraffic']}'"
        )

    if "RoutePolling" in props:
        route_polling_val = string_to_boolean(props["RoutePolling"])
        if route_polling_val is None:
            wg_cn_errors.append(f"RoutePolling misconfigured, must be True or False")
        else:
            props["RoutePolling"] = route_polling_val
    else:
        props["RoutePolling"] = LibConfig.Defaults.ROUTE_POLLING
        logger.debug(
            f"RoutePolling not configured, can be True or False, setting to '{props['RoutePolling']}'"
        )

    if "RoutePollingInterval" in props:
        interval_val = calc_string_seconds(props["RoutePollingInterval"])
        if interval_val is None:
            props["RoutePollingInterval"] = LibConfig.Defaults.POLLING_INTERVAL
            logger.debug(
                "RoutePollingInterval is not specified, using default: %s"
                % LibConfig.Defaults.POLLING_INTERVAL
            )
        elif interval_val < LibConfig.MIN_POLLING_INTERVAL:
            props["RoutePollingInterval"] = LibConfig.MIN_POLLING_INTERVAL
            logger.debug(
                "RoutePollingInterval less than minimum value. Using min: %s"
                % LibConfig.MIN_POLLING_INTERVAL
            )
        else:
            props["RoutePollingInterval"] = interval_val
    else:
        # Add default if DNE
        props["RoutePollingInterval"] = LibConfig.Defaults.POLLING_INTERVAL
        logger.debug(
            f"RoutePollingInterval not configured, setting to '{props['RoutePollingInterval']}' seconds"
        )

    if "HandshakeTimeout" in props:
        timeout_val = calc_string_seconds(props["HandshakeTimeout"])
        if timeout_val is None:
            wg_cn_errors.append(
                "HandshakeTimeout misconfigured, must be integer ending with 's, m, or h'"
            )
        elif timeout_val < LibConfig.MIN_HANDSHAKE_TIMEOUT:
            props["HandshakeTimeout"] = LibConfig.MIN_HANDSHAKE_TIMEOUT
            logger.debug(
                "HandshakeTimeout less than minimum value. Using min: %s"
                % LibConfig.MIN_HANDSHAKE_TIMEOUT
            )
        else:
            props["HandshakeTimeout"] = timeout_val
    else:
        # Add default if DNE.
        props["HandshakeTimeout"] = LibConfig.Defaults.HANDSHAKE_TIMEOUT
        logger.debug(
            f"HandshakeTimeout not configured, setting to '{props['HandshakeTimeout']}'"
        )

    if "LocalRoutes" in props and props["LocalRoutes"] != "":
        for route in props["LocalRoutes"].split(","):
            if not re.match(
                wireguard.CIDR_CHECK_REGEX,
                route.strip(),
            ):
                wg_cn_errors.append(f"Local route not in CIDR format: {route}")
    else:
        props["LocalRoutes"] = ""
        logger.debug(f"No local routes specified in LocalRoutes")

    if "SecondaryEndpoints" in props and props["SecondaryEndpoints"] == "":
        logger.debug(
            f"No secondary controller endpoints specified in SecondaryEndpoints"
        )
    elif "SecondaryEndpoints" not in props:
        props["SecondaryEndpoints"] = ""
        logger.debug(f"SecondaryEndpoints not configured, assuming None")

    if validate_endpoints:
        valid_endpoint, endpoint_errors = get_first_valid_endpoint(
            get_secondary_endpoints_props(props), logger
        )

        if valid_endpoint:
            props = update_props_endpoint(props, valid_endpoint)
        if endpoint_errors:
            wg_cn_errors = wg_cn_errors + endpoint_errors

    errors = {}
    if wg_native_errors:
        errors[ConnectionTypes.WIREGUARD_NATIVE] = wg_native_errors
    if wg_cn_errors:
        errors[ConnectionTypes.WIREGUARD_COHESIVE] = wg_cn_errors

    logger.debug(f"Clientpack properties validated")
    return errors


def do_route_polling(props):
    if props["type"] == ConnectionTypes.WIREGUARD_NATIVE:
        return False
    if props.get("TunnelAllTraffic", LibConfig.Defaults.TUNNEL_ALL_TRAFFIC):
        lib_logger.debug("Tunnel all traffic enabled, no route polling required")
        return False
    elif props.get("RoutePolling", LibConfig.Defaults.ROUTE_POLLING):
        lib_logger.debug("Route polling enabled")
        return True
    return False


def tunnel_all_traffic(props):
    if props["type"] == ConnectionTypes.WIREGUARD_NATIVE:
        return False
    return props.get("TunnelAllTraffic", LibConfig.Defaults.TUNNEL_ALL_TRAFFIC)


def get_first_valid_endpoint(endpoints, logger=None, timeout=None):
    """get_first_valid_endpoint

    Args:
        endpoints ([List|str]): list of strings or string csv of endpoints (ip,dns)
        logger (optional): Defaults to library logger.

    Returns:
        tuple: (endpoint [str], errors [List[str]])
    """
    errors = []

    if not endpoints:
        # return empty
        return (None, errors)

    logger = logger or lib_logger
    errors = []
    if type(endpoints) is str:
        endpoints_cln = [x.strip() for x in endpoints.split(",")]
    else:
        endpoints_cln = [x.strip() for x in endpoints]

    for ep in endpoints_cln:
        logger.debug(f"Connecting to VNS3 controller endpoint: {ep}")
        try:
            requests.get(
                f"http://{ep}:8000",
                timeout=(timeout or LibConfig.DEFAULT_CONNECT_TIMEOUT),
                verify=False,
            )
            logger.info(f"Connected to VNS3 controller endpoint: {ep}")
            return (ep, errors)
        except:
            logger.info(f"Failed to connect to VNS3 controller endpoint: {ep}")
    errors.append(
        f"Could not connect to any controller endpoints specified: {endpoints}"
    )
    return (None, errors)


def handle_signal_exit(env, *args, **kwargs):
    env.logger.debug("Received exit signal")
    sys.exit(0)


def teardown_connection(env, client_props):
    """teardown_connection

    Run teardown steps for VPN connection

    Args:
        env (config.EnvConfig):
        client_props (dict): required properties:
            ActiveInterface: str,
            wg_conf_file: str
            type: str,
            ClientpackId: str,
            Secret: str,
            openid: str,
            refresh_token: str,
            Endpoint: str

    Returns:
        _type_: _description_
    """
    logger = env.logger or lib_logger
    logger.info(f"Shutting down Wireguard interface {client_props['ActiveInterface']}")
    client_interface = client_props["ActiveInterface"]

    try:
        wireguard.wg_quick_down(env, client_interface, client_props["wg_conf_file"])
    except Exception as e:
        traceback.print_exc(file=sys.stdout)
        logger.error(
            "Shutdown interface down error. Please run as admin/sudo. Error: %s"
            % str(e)
        )

    if not wireguard.is_interface_down(env, client_interface):
        logger.error(
            (
                "Failed to down interface %s for connection %s"
                % (client_interface, client_props.get("name", ""))
            ).strip()
        )
        return False

    if client_props["type"] == ConnectionTypes.WIREGUARD_COHESIVE:
        try:
            wg_disconnect(
                get_current_endpoint(client_props),
                client_props["ClientpackId"],
                client_props["Secret"],
                logger=logger,
            )
            delete_all_routes(env, client_props)

        except Exception as e:
            traceback.print_exc(file=sys.stdout)
            logger.error("Shutdown disconnect error: %s" % str(e))
            logger.debug(traceback.format_exc())

    keys_to_remove = ["pid", "runtime", "token", "openid", "refresh_token"]
    if env.is_macos:
        keys_to_remove.append("interface")

    state_file = env.state_file or client_props.get("state_file")
    if state_file:
        file_utils.update_data_file(
            state_file,
            {
                "status": "disconnected",
                "last_disconnect": datetime.utcnow().strftime(
                    LibConfig.DATETIME_FORMAT
                ),
            },
            remove_keys=keys_to_remove,
        )
    logger.info("==== Connection teardown completed ====")
    return True


def exit_handler(env, props, *args, **kwargs):
    env.logger.debug("Exit handler called for connection %s" % props["name"])
    teardown_connection(env, props)
    sys.exit(0)


def string_to_boolean(val):
    if type(val) is bool:
        return val
    if val == "":
        return None
    val_lower = val.lower()
    if not any([val_lower == "true", val_lower == "false"]):
        return None
    return val_lower == "true"


def calc_string_seconds(val, default_unit="s"):
    """calc_string_seconds

    Args:
        val (str): string of form [digits][s,m,h]
        default_unit (str, optional): Defaults to "s".

    Returns:
        None or integer: if None is returned, string is invalid.
    """
    if val == "":
        return None
    elif type(val) in [int, float]:
        return int(val)

    # matches [integer](s,m,h) where s,m,h optional.
    parts_match = re.search(r"([0-9]*)([smh]{0,1})", val)
    if not parts_match:
        return None

    number = parts_match[1]
    matched_unit = parts_match[2]

    if not matched_unit:
        if len(number) == len(val):
            # ie number.isdigit() is True
            unit = default_unit
        else:
            # indicates invalid string
            return None
    else:
        unit = matched_unit

    seconds_per_unit = {"s": 1, "m": 60, "h": 3600}
    if unit not in seconds_per_unit:
        return None

    return int(number) * seconds_per_unit[unit]


def build_client_props(clientpack_file, conn_data):
    props = parse_clientpack(clientpack_file)
    props["name"] = conn_data["name"]
    props["wg_conf_file"] = props["wg_cp_file"]
    props["state_file"] = conn_data.get("state_file")
    props["routes_file"] = conn_data.get("routes_file")
    props["type"] = conn_data.get("type") or ConnectionTypes.DEFAULT
    return props


def run_client(env, params, should_thread_stop=None):
    """run_client

    Blocking process. Build WG connection to VNS3.

    Args:
        env (config.EnvConfig)
        params (dict): {
            clientpack: str, path to clientpack file,
            routes: file path, (required),
            state_file: file path, path to write state key:value pairs to
            interface: expected interface to use.
            log_file: file path, library output if None (which defaults to stdout),
            log_level: str, override env log level for file output
        }

    Raises:
        ClientException: str
    """
    routes_file = params.get("routes_file")
    wg_cp_file = params.get("clientpack")

    state_file = params.get("state_file")
    log_file = params.get("log_file")
    params_logger = params.get("logger")
    interface = params.get("interface")
    conn_type = params.get("type") or ConnectionTypes.DEFAULT

    running_in_thread = should_thread_stop is not None
    if not running_in_thread:
        should_thread_stop = lambda: False

    log_level = params.get("log_level", env.get("log_level")).upper()
    if params_logger:
        logger = params_logger
    elif log_file:
        logger = configure_client_logger(
            params["name"], log_level, params.get("log_file"), env.get("log_format")
        )
    else:
        logger = lib_logger

    env.state_file = state_file
    env.logger = logger

    # if not env.is_admin:
    #     error = "Permissions error: must run as administrator"
    #     logger.error(error)
    #     raise ClientException(error)

    if not wg_cp_file:
        error = "Clientpack (conf file) is required for run client"
        logger.error(error)
        raise ClientException(error)

    if not os.path.exists(wg_cp_file):
        error = "Clientpack file not found %s" % wg_cp_file
        logger.error(error)
        raise ClientException(error)

    logger.debug("Starting client [params=%s]" % json_stringify(params))

    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
    # NOTE: wg_cp_file is a clientpack! wg_conf_file is Wireguard conf file!
    # Question: Should we entirely remove the default and enforce clientpack param
    # is passed along?
    # parse clientpack tokens from file
    props = build_client_props(wg_cp_file, params)
    # validation mutates props
    validation_errors = validate_clientpack(props, logger)
    if validation_errors:
        native_wg_errors = validation_errors.get(ConnectionTypes.WIREGUARD_NATIVE, [])
        wg_cn_errors = validation_errors.get(ConnectionTypes.WIREGUARD_COHESIVE, [])

        # any errors fail for CN type
        if conn_type == ConnectionTypes.WIREGUARD_COHESIVE:
            error = "Problems were found starting up wireguard client: %s" % (
                native_wg_errors + wg_cn_errors
            )
            logger.error(error)
            raise ClientException(error)
        else:
            if len(native_wg_errors) > 0:
                error = (
                    "Problems were found starting up native wireguard: %s"
                    % native_wg_errors
                )
                logger.error(error)
                raise ClientException(error)

    with_route_polling = do_route_polling(props)
    if with_route_polling and not routes_file:
        error = "routes_file param is required if conf is configured for route polling"
        logger.error(error)
        raise ClientException(error)

    # copy clientpack to wg dir
    initialize(env, props)

    state_updates = {}
    if running_in_thread:
        state_updates["runtime"] = "thread"
        state_updates["pid"] = threading.get_ident()
    else:
        state_updates["runtime"] = "subprocess"
        state_updates["pid"] = os.getpid()

    # Notes about interfaces
    # - Wireguard expects conf file to be named [interface].conf
    # - Mac does not respect interface name but instead creates next available
    #    utun interface (even if conf file is utun5.conf). Linux and Windows
    #    seem to create interfaces by same name as conf file.
    # So for mac basically we have to "find" the interface created for a particular
    # conf file. we do this based on overlay IP (find_ip_interface)

    # if interface not passed explicitly, use clientpack name
    if not interface:
        # use interface parsed from clientpack name
        interface = props["interface"]

    if env.is_macos:
        props["ActiveInterface"] = None
    else:
        props["ActiveInterface"] = interface
        state_updates.update(interface=interface)

    exit_handler_bound = partial(handle_signal_exit, env)
    if not running_in_thread:
        signal_code = signal.SIGBREAK if env.is_windows else signal.SIGTERM
        signal.signal(signal_code, exit_handler_bound)
        atexit.register(exit_handler, env, props)

    file_utils.update_data_file(env.state_file, state_updates)

    default_client_monitor_interval = env.get("conn_monitor_interval")
    login_retry_interval = env.get("login_retry_interval")
    polling_interval = props["RoutePollingInterval"]
    handshake_timeout_interval = props["HandshakeTimeout"]
    logger.debug("Starting main process loop with properties: %s" % props)
    try:
        running_native_only = props["type"] == ConnectionTypes.WIREGUARD_NATIVE
        if running_native_only:
            logger.info("Running native wireguard.")
        else:
            # generate wg config file
            update_wg_config(env, props, logger=env.logger)

            logger.info(
                "Wireguard interface will be reset if no handshake within %s seconds"
                % handshake_timeout_interval
            )

            if with_route_polling:
                logger.debug(
                    f"Route polling enabled, will be synchronizing routes every {props['RoutePollingInterval']} seconds"
                )
            else:
                logger.debug(
                    f"Route polling disabled, will not be synchronizing routes every {props['RoutePollingInterval']}"
                )

        # Question: should we fail IF there is already a running wg process
        # for the same interface? WG behavior is to down any wg interface with
        # the same name even if its a totally different conf file.
        # So we would have to determine that the current running interface
        # was NOT the one as configured by this clientpack. Perhaps by looking
        # at pass clientpacks Address, and finding IP for the interface.
        # Note: this would not be possible on mac
        wireguard.wg_quick_down(env, props["ActiveInterface"], props["wg_conf_file"])

        while True:
            display_interface_status(
                env, props["ActiveInterface"], handshake_timeout_interval
            )
            if needs_interface_reset(env, props, handshake_timeout_interval):
                if not running_native_only:
                    delete_local_routes(env, props)
                    add_local_routes(env, props, logger=env.logger)
                    if tunnel_all_traffic(props):
                        add_controller_routes(env, props)
                reset_interface(env, props["ActiveInterface"], props, logger=env.logger)

            if props["ConnectionStatus"] != "up":
                env.logger.debug(
                    "Connection not up. Retry again in %s seconds"
                    % login_retry_interval
                )
                time.sleep(login_retry_interval)
            elif with_route_polling:
                update_routes(
                    env,
                    props,
                    props["routes_file"],
                    props["controller_vip"],
                    props["ActiveInterface"],
                    props["ClientpackId"],
                    props["Address"],
                )
                time.sleep(polling_interval)
            else:
                time.sleep(default_client_monitor_interval)

            if should_thread_stop():
                env.logger.info("Thread stop request received. Shutting down.")
                break

        # if it gets here, we are exiting!
        exit_handler(env, props)
    except Exception as e:
        env.logger.error("Fatal exception for client: %s" % str(e))
        if log_level == "DEBUG":
            traceback.print_exc(file=sys.stdout)
            env.logger.debug(traceback.format_exc())

        if running_in_thread:
            exit_handler(env, props)
        else:
            sys.exit(1)
